@@ -16,7 +16,7 @@ import usb
 from datetime import datetime
 from enum import IntEnum, IntFlag
 
-from amaranth                            import Signal, Elaboratable, Module, DomainRenamer, ResetInserter, C, Mux
+from amaranth                            import Signal, Elaboratable, Module, DomainRenamer, ResetInserter, C, Mux, Array
 from amaranth.build.res                  import ResourceError
 from usb_protocol.emitters               import DeviceDescriptorCollection
 from usb_protocol.types                  import USBRequestType, USBRequestRecipient
@@ -30,6 +30,7 @@ from luna.gateware.stream.generator      import StreamSerializer
 from luna.gateware.architecture.car      import LunaECP5DomainGenerator
 from luna.gateware.architecture.flash_sn import ECP5FlashUIDStringDescriptor
 from luna.gateware.interface.ulpi        import UTMITranslator
+from luna.gateware.usb.usb2              import USBSpeed
 from luna.gateware.usb.usb2.control      import USBControlEndpoint
 from luna.gateware.usb.request.standard  import StandardRequestHandler
 from luna.gateware.usb.request.windows   import MicrosoftOS10DescriptorCollection, MicrosoftOS10RequestHandler
@@ -41,14 +42,14 @@ from usb_protocol.types.descriptors.microsoft10 import RegistryTypes
 
 from .analyzer                           import USBAnalyzer
 from .fifo                               import Stream16to8, StreamFIFO, AsyncFIFOReadReset, HyperRAMPacketFIFO
+from .speed_detection                    import USBAnalyzerSpeedDetector
+from .event_detection                    import USBHighSpeedEventDetector, USBFullSpeedEventDetector, USBLowSpeedEventDetector
+from .speeds                             import USBAnalyzerSpeed
+from .events                             import USBAnalyzerEvent
 
 import cynthion
 
 
-USB_SPEED_HIGH       = 0b00
-USB_SPEED_FULL       = 0b01
-USB_SPEED_LOW        = 0b10
-
 USB_VENDOR_ID        = cynthion.shared.usb.bVendorId.cynthion
 USB_PRODUCT_ID       = cynthion.shared.usb.bProductId.cynthion
 
@@ -184,6 +185,12 @@ class USBAnalyzerVendorRequestHandler(ControlRequestHandler):
                         USBAnalyzerSupportedSpeeds.USB_SPEED_LOW | \
                         USBAnalyzerSupportedSpeeds.USB_SPEED_FULL | \
                         USBAnalyzerSupportedSpeeds.USB_SPEED_HIGH
+
+                    # Automatic speed detection is only supported on Cynthion r0.6+.
+                    if platform.version >= (0, 6):
+                        supported_speeds |= \
+                            USBAnalyzerSupportedSpeeds.USB_SPEED_AUTO
+
                     self.handle_simple_data_request(m, transmitter, supported_speeds, length=1)
 
                 # SET_TEST_CONFIG -- The host is trying to configure our test device
@@ -257,6 +264,7 @@ class USBAnalyzerApplet(Elaboratable):
 
         # State register
         m.submodules.state = state = USBAnalyzerRegister()
+        speed_selection = state.current[USBAnalyzerState.SPEED]
 
         # Test config register
         m.submodules.test_config = test_config = USBAnalyzerRegister(reset=0x01)
@@ -269,6 +277,21 @@ class USBAnalyzerApplet(Elaboratable):
         ulpi = platform.request("target_phy")
         m.submodules.utmi = utmi = UTMITranslator(ulpi=ulpi)
 
+        # Add event detectors for fixed speeds.
+        m.submodules.hs_event = hs_event_detector = USBHighSpeedEventDetector()
+        m.submodules.fs_event = fs_event_detector = USBFullSpeedEventDetector()
+        m.submodules.ls_event = ls_event_detector = USBLowSpeedEventDetector()
+        m.d.comb += [
+            hs_event_detector.reset.eq(state.write),
+            fs_event_detector.reset.eq(state.write),
+            ls_event_detector.reset.eq(state.write),
+            fs_event_detector.line_state.eq(utmi.line_state),
+            ls_event_detector.line_state.eq(utmi.line_state),
+            hs_event_detector.vbus_connected.eq(utmi.session_valid),
+            fs_event_detector.vbus_connected.eq(utmi.session_valid),
+            ls_event_detector.vbus_connected.eq(utmi.session_valid),
+        ]
+
         # Connect our power controls. The power_control_enable bit must be set
         # to use this feature, otherwise the default pass-through is enabled.
         power_control_enable = state.current[USBAnalyzerState.POWER_CONTROL_ENABLE]
@@ -290,6 +313,39 @@ class USBAnalyzerApplet(Elaboratable):
                     Mux(power_control_enable,
                         state.current[USBAnalyzerState.VBUS_TARGET_A_DISCHARGE], False)),
             ]
+
+            # Tap the D+/D- signals for speed detection.
+            usb_dp = Signal()
+            usb_dm = Signal()
+            usb_dp_input = platform.request("target_usb_dp_chirp").i
+            usb_dm_input = platform.request("target_usb_dm_chirp").i
+            m.d.usb += [
+                usb_dp.eq(usb_dp_input),
+                usb_dm.eq(usb_dm_input),
+            ]
+
+            # Add a speed detector and use it when selected.
+            m.submodules.speed = speed_detector = USBAnalyzerSpeedDetector()
+            phy_speed = Mux(
+                speed_selection == USBAnalyzerSpeed.AUTO,
+                speed_detector.phy_speed,
+                speed_selection)
+            detected_speed = Mux(
+                speed_selection == USBAnalyzerSpeed.AUTO,
+                speed_detector.detected_speed,
+                speed_selection)
+            auto_event_strobe = speed_detector.event_strobe
+            auto_event_code = speed_detector.event_code
+
+            # Provide the necessary signals for speed detection.
+            m.d.comb += [
+                speed_detector.reset.eq(state.write),
+                speed_detector.line_state.eq(utmi.line_state),
+                speed_detector.usb_dp.eq(usb_dp),
+                speed_detector.usb_dm.eq(usb_dm),
+                speed_detector.vbus_connected.eq(utmi.session_valid),
+            ]
+
         else:
             m.d.comb += [
                 # On the r0.1 to r0.5 boards, power switching is different.
@@ -309,16 +365,38 @@ class USBAnalyzerApplet(Elaboratable):
                 # port, and no discharge capability on TARGET-A.
             ]
 
+            # Speed selection is manual only.
+            phy_speed = detected_speed = next_speed = speed_selection
+            auto_event_strobe = False
+            auto_event_code = USBAnalyzerEvent.NONE
+
+        # Choose the appropriate event source according to speed selection.
+        event_strobes = Array([
+            hs_event_detector.event_strobe,
+            fs_event_detector.event_strobe,
+            ls_event_detector.event_strobe,
+            auto_event_strobe,
+        ])
+
+        event_codes = Array([
+            hs_event_detector.event_code,
+            fs_event_detector.event_code,
+            ls_event_detector.event_code,
+            auto_event_code,
+        ])
+
+        event_strobe = event_strobes[speed_selection]
+        event_code = event_codes[speed_selection]
+
         # Set up our parameters.
         m.d.comb += [
-
-            # Set PHY mode to non-driving as we want to passively observe.
+            # Set PHY mode to non-driving and to the desired speed.
             #
             # `dp_pulldown`, `dm_pulldown` and `term_select` do not need to be
             # configured as these values are "don't cares" for this specific
             # `op_mode` (see ULPI Specification rev. 1.1 Table 41).
             utmi.op_mode     .eq(0b01),
-            utmi.xcvr_select .eq(state.current[USBAnalyzerState.SPEED]),
+            utmi.xcvr_select .eq(phy_speed),
         ]
 
         # Select the appropriate PHY according to platform version.
@@ -381,10 +459,11 @@ class USBAnalyzerApplet(Elaboratable):
         usb.add_endpoint(stream_ep)
 
         # Create a USB analyzer.
-        m.submodules.analyzer = analyzer = USBAnalyzer(utmi_interface=utmi)
+        m.submodules.analyzer = analyzer = USBAnalyzer(
+            utmi, utmi.session_valid, detected_speed, event_strobe, event_code)
 
         # Follow this with a HyperRAM FIFO for additional buffering.
-        reset_on_start = ResetInserter(analyzer.discarding)
+        reset_on_start = ResetInserter(analyzer.starting)
         m.submodules.psram_fifo = psram_fifo = reset_on_start(
             HyperRAMPacketFIFO(out_fifo_depth=128))
 
@@ -402,14 +481,14 @@ class USBAnalyzerApplet(Elaboratable):
             # Flush endpoint when analyzer is idle with capture disabled.
             stream_ep.flush             .eq(analyzer.idle & ~analyzer.capture_enable),
 
-            # Discard data buffered by endpoint when the analyzer discards its data.
-            stream_ep.discard           .eq(analyzer.discarding),
+            # Discard old data buffered by endpoint when the analyzer starts.
+            stream_ep.discard           .eq(analyzer.starting),
 
             # USB stream pipeline.
             psram_fifo.input            .stream_eq(analyzer.stream),
             s16to8.input                .stream_eq(psram_fifo.output),
             clk_conv.input              .stream_eq(s16to8.output),
-            clk_conv.fifo.ext_rst       .eq(analyzer.discarding),
+            clk_conv.fifo.ext_rst       .eq(analyzer.starting),
             stream_ep.stream            .stream_eq(clk_conv.output),
 
             usb.connect                 .eq(1),
@@ -431,24 +510,24 @@ class USBAnalyzerApplet(Elaboratable):
 class AnalyzerTestDevice(Elaboratable):
     """ Built-in example device that can be used to test the analyzer. """
 
-    SPEEDS = (USB_SPEED_HIGH, USB_SPEED_FULL, USB_SPEED_LOW)
+    SPEEDS = (USBSpeed.HIGH, USBSpeed.FULL, USBSpeed.LOW)
 
     EP0_MAX_SIZE = {
-        USB_SPEED_HIGH: 64,
-        USB_SPEED_FULL: 64,
-        USB_SPEED_LOW: 8,
+        USBSpeed.HIGH: 64,
+        USBSpeed.FULL: 64,
+        USBSpeed.LOW: 8,
     }
 
     INT_EP_MAX_SIZE = {
-        USB_SPEED_HIGH: 512,
-        USB_SPEED_FULL: 64,
-        USB_SPEED_LOW: 8,
+        USBSpeed.HIGH: 512,
+        USBSpeed.FULL: 64,
+        USBSpeed.LOW: 8,
     }
 
     INT_EP_NUM = {
-        USB_SPEED_HIGH: 1,
-        USB_SPEED_FULL: 2,
-        USB_SPEED_LOW: 3,
+        USBSpeed.HIGH: 1,
+        USBSpeed.FULL: 2,
+        USBSpeed.LOW: 3,
     }
 
     def __init__(self, config):
@@ -488,8 +567,8 @@ class AnalyzerTestDevice(Elaboratable):
         current_speed = self.config.current[1:3]
         m.d.comb += [
             usb.connect.eq(self.config.current[0]),
-            usb.low_speed_only.eq(current_speed == USB_SPEED_LOW),
-            usb.full_speed_only.eq(current_speed == USB_SPEED_FULL),
+            usb.low_speed_only.eq(current_speed == USBSpeed.LOW),
+            usb.full_speed_only.eq(current_speed == USBSpeed.FULL),
         ]
 
         # Create control endpoint.
